from unittest import mock

import pytest
from pyannotate_runtime.collect_types import resolve_type

from ntgen.models import Attribute
from ntgen.models import NT
from ntgen.utils import normalize_field_name


@pytest.fixture()
def sample_person_nt():
    return NT(
        attrs=[
            Attribute(original_name="name", type=str, value="John"),
            Attribute(original_name="age", type=int, value=24),
            Attribute(
                original_name="address",
                type=NT(attrs=[Attribute("city", str, "Warsaw"), Attribute("street", str, "Biała 13")], name="address",),
                value={"city": "Warsaw", "street": "Biała 13"},
            ),
        ],
        name="person",
    )


class TestAttribute:
    @pytest.fixture(
        params=[None, 1, "string", [1, 2, 3], {1: "1"}, pytest.lazy_fixture("sample_person_nt"),]
    )
    def sample_attribute(self, request, sample_field_name):
        return Attribute(
            original_name=sample_field_name,
            type=request.param if isinstance(request.param, NT) else resolve_type(request.param),
            value=request.param,
        )

    def test_repr_field_name(self, sample_attribute):
        with mock.patch("ntgen.models.normalize_field_name", side_effect=normalize_field_name,) as mock_normalize_field_name:
            assert sample_attribute.repr_field_name == normalize_field_name(
                name=sample_attribute.original_name, leading_undescores_prefix=None,
            )
            mock_normalize_field_name.assert_called_once()

    @pytest.mark.parametrize(
        ("sample_attribute", "expected_type_hint"),
        (
            (1, "int"),
            ("string", "str"),
            (None, "None"),
            ({1: "1"}, "Dict[int, str]"),
            (pytest.lazy_fixture("sample_person_nt"), "Person"),
        ),
        indirect=["sample_attribute"],
    )
    def test_repr_type_hint(self, sample_attribute: Attribute, expected_type_hint: str):
        assert sample_attribute.repr_type_hint == expected_type_hint

    @pytest.mark.parametrize(
        ("sample_attribute", "is_user_defined"),
        ((1, False), ("string", False), (None, False), ({1: "1"}, False), (pytest.lazy_fixture("sample_person_nt"), True),),
        indirect=["sample_attribute"],
    )
    def test_is_user_defined(self, sample_attribute: Attribute, is_user_defined: str):
        assert sample_attribute.is_user_defined == is_user_defined


class TestNT:
    def test_attr_views(self, sample_person_nt: NT):
        assert sample_person_nt.builtin_type_attrs == [
            Attribute(original_name="name", type=str, value="John"),
            Attribute(original_name="age", type=int, value=24),
        ]
        assert sample_person_nt.nt_attrs == [
            Attribute(
                original_name="address",
                type=NT(attrs=[Attribute("city", str, "Warsaw"), Attribute("street", str, "Biała 13")], name="address",),
                value={"city": "Warsaw", "street": "Biała 13"},
            )
        ]



from typing import Dict

import pytest

from ntgen.utils import normalize_field_name
from ntgen.utils import replace_leading_underscores
from tests.sample_dicts import APARTMENT_DICT
from tests.sample_dicts import EX_DICT


@pytest.fixture(params=[(APARTMENT_DICT, "akelius_dict"), (EX_DICT, "example_dict")])
def sample_dict(request):
    return request.param


@pytest.fixture(params=["sample_name", "__sample_name", "SampleName", "sample-name"])
def sample_field_name(request):
    return request.param


def normalize_keys(data: Dict) -> Dict:
    assert isinstance(data, dict)

    normalized_dict = {}
    for key, value in data.items():
        normalized_key = replace_leading_underscores(normalize_field_name(key))
        normalized_dict[normalized_key] = normalize_keys(value) if isinstance(value, dict) else value
    return normalized_dict



import importlib
from pathlib import Path
from shutil import copyfile

from ntgen import generate_from_dict
from ntgen.utils import normalize_class_name
from tests.conftest import normalize_keys


class TestGeneratedDictionaryIsEquivalentWithOriginal:
    module_template_path = Path("tests") / "end_to_end" / "module_template.py"

    def test_interpreted_dict_exports_to_the_same_as_input(self, tmp_module, sample_dict):
        example_dict_object, submodule_name = sample_dict
        module_path = copyfile(self.module_template_path, tmp_module / f"{submodule_name}.py")

        main_object_name = normalize_class_name(submodule_name)
        nt_code = generate_from_dict(data=example_dict_object, name=main_object_name, insert_as_dict=True, insert_from_dict=True)

        with module_path.open("a") as f:
            f.write(nt_code)

        test_module_obj = importlib.import_module(str(tmp_module / submodule_name).replace("/", "."))
        nt_class = getattr(test_module_obj, main_object_name)

        nt_object = nt_class.from_dict(example_dict_object)
        nt_exported_to_dict = nt_object.as_dict()

        assert normalize_keys(nt_exported_to_dict) == normalize_keys(example_dict_object)



from ntgen.config import Config
from ntgen.models import Attribute
from ntgen.models import NT
from ntgen.utils import indent_statement


def _render_subtuples(nt_definition: NT, config: Config) -> str:
    return "".join(
        _render_nt_definition(sub_tuple.type, config=config) for sub_tuple in nt_definition.attrs if isinstance(sub_tuple.type, NT)
    )


def _render_signature(nt_definition: NT) -> str:
    return f"class {nt_definition}(NamedTuple):\n"


def _render_attribute(attr: Attribute) -> str:
    return f"{attr.repr_field_name}: {attr.repr_type_hint}\n"


def _render_attributes(nt_definition: NT) -> str:
    return "".join(indent_statement(indent=1, statement=_render_attribute(attr)) for attr in nt_definition.attrs)


def _render_constructor(nt_definition: NT) -> str:
    param_assignments = []
    for attr in nt_definition.builtin_type_attrs:
        param_assignments.append(f"{attr.repr_field_name}=data['{attr.original_name}'],")

    for attr in nt_definition.nt_attrs:
        param_assignments.append(f"{attr.repr_field_name}={attr.type}.from_dict(data['{attr.original_name}']),")

    indents_statements = (
        (1, "@classmethod"),
        (1, f"def from_dict(cls, data: Dict) -> {nt_definition}:"),
        (2, f"return {nt_definition}("),
        *[(3, param) for param in param_assignments],
        (2, ")"),
    )

    return "\n".join(indent_statement(indent, statement) for indent, statement in indents_statements) + "\n\n"


def _render_as_dict_method() -> str:
    indents_statements = (
        (1, f"def as_dict(self) -> Dict:"),
        (2, f"return {{"),
        (3, f'name: (value.as_dict() if hasattr(value, "as_dict") else value)'),
        (3, f"for name, value in self._asdict().items()"),
        (2, f"}}"),
    )

    return "\n".join(indent_statement(indent, statement) for indent, statement in indents_statements) + "\n\n"


def _render_nt_definition(nt_definition: NT, config: Config) -> str:
    return "".join(
        (
            _render_subtuples(nt_definition, config=config),
            "\n",
            _render_signature(nt_definition),
            _render_attributes(nt_definition),
            "\n",
            _render_constructor(nt_definition) if config.insert_from_dict else "",
            _render_as_dict_method() if config.insert_as_dict else "",
        ),
    )


def render_nt_definition(nt_definition: NT, config: Config) -> str:
    """
    Render a NamedTuple type class definition based on a given NT object.

    :param nt_definition: NT object
    :param config: Config object
    :return: string containing valid Python code declaring a NamedTuple
    """
    return _render_nt_definition(nt_definition=nt_definition, config=config).lstrip().rstrip()



"""A module for storing larger test data in Python format."""
from typing import Any
from typing import Dict

APARTMENT_DICT: Dict[str, Any] = {
    "id": "1234-1234",
    "type": "living",
    "countryCode": "DE",
    "availableUntilDate": None,
    "availableFromDate": "2018-05-16T00:00:00Z",
    "availableFromNowOn": False,
    "publishedFromDate": None,
    "address": {
        "borough": "Dulsberg",
        "city": "Hamburg",
        "houseNumber": "2",
        "latitude": 53.587485,
        "longitude": 10.063215,
        "postalCode": "22049",
        "streetName": "Nordschleswiger Straße",
        "area": "Hamburg",
    },
    "_attachments": "attachments/",
    "_ts": 15828103462,
}

EX_DICT: Dict[str, Any] = {
    "butter": 3,
    "name": "Thomas",
    "volume": 3.14,
    "items": [1, 2, 3],
    "obligatory_items": (1, 2, 3),
}



from __future__ import annotations

from typing import NamedTuple
from typing import Optional


class Config(NamedTuple):
    """NamedTuple class for storage of named tuple rendering options."""

    snake_case: bool = True
    camel_case: bool = True
    insert_from_dict: bool = False
    insert_as_dict: bool = False
    max_level: Optional[int] = None



from unittest import mock

import pytest

from ntgen.utils import convert_to_snake_case
from ntgen.utils import indent_statement
from ntgen.utils import normalize_class_name
from ntgen.utils import normalize_field_name
from ntgen.utils import replace_leading_underscores


class TestUtils:
    @pytest.fixture(params=[None, "sample_prefix", "another"])
    def sample_prefix(self, request):
        return request.param

    @pytest.fixture(params=["SamplePascalName123", "sample_snake_case_name_123", "sample-kebab-case-123"])
    def sample_name(self, request):
        return request.param

    @pytest.mark.parametrize(
        ("indent", "statement", "expected_string"),
        (
            (0, "some_statement", "some_statement"),
            (1, "some_statement", "    some_statement"),
            (2, "some_statement", "        some_statement"),
            (0, "def test():", "def test():"),
            (1, "def test():", "    def test():"),
            (2, "def test():", "        def test():"),
        ),
    )
    def test_indent_statement(self, indent, statement, expected_string):
        assert indent_statement(indent=indent, statement=statement) == expected_string

    def test_normalize_field_name(self, sample_name, sample_prefix):
        with mock.patch("ntgen.utils.convert_to_snake_case",) as mock_convert_to_snake_case, mock.patch(
            "ntgen.utils.replace_leading_underscores",
        ) as mock_replace_leading_underscores:
            normalize_field_name(name=sample_name, leading_undescores_prefix=sample_prefix)

            mock_convert_to_snake_case.assert_called_once()
            mock_replace_leading_underscores.assert_called_once()

    @pytest.mark.parametrize(
        ("name", "expected_output"),
        (
            ("SamplePascalName123", "sample_pascal_name123"),
            ("sample_snake_case_name_123", "sample_snake_case_name_123"),
            ("sample-kebab-case-123", "sample_kebab_case_123"),
        ),
    )
    def test_convert_to_snake_case(self, name, expected_output):
        assert convert_to_snake_case(name) == expected_output

    @pytest.mark.parametrize(
        ("name", "expected_output"),
        (
            ("SamplePascalName123", "SamplePascalName123"),
            ("sample_snake_case_name_123", "SampleSnakeCaseName123"),
            ("sample-kebab-case-123", "SampleKebabCase123"),
        ),
    )
    def test_normalize_class_name(self, name, expected_output):
        assert normalize_class_name(name) == expected_output

    @pytest.mark.parametrize(
        ("name", "prefix", "expected_output"),
        (
            ("normalized_name", None, "normalized_name"),
            ("normalized_name", "some_prefix", "normalized_name"),
            ("_incorrect_name", None, "incorrect_name"),
            ("__incorrect_name", None, "incorrect_name"),
            ("__incorrect_name", "some_prefix", "some_prefix_incorrect_name"),
            ("____incorrect_name", "some_prefix", "some_prefix_incorrect_name"),
        ),
    )
    def test_replace_leading_underscores(self, name, prefix, expected_output):
        assert replace_leading_underscores(name, prefix=prefix) == expected_output



from pathlib import Path
from shutil import rmtree

import pytest


@pytest.fixture(autouse=True)
def tmp_module():
    tmp_module = Path("tests") / "tmp_module"

    tmp_module.mkdir()
    (tmp_module / "__init__.py").touch()

    yield tmp_module

    rmtree(tmp_module, ignore_errors=True)



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Type
from typing import Union

from pyannotate_runtime.collect_types import InternalType
from pyannotate_runtime.collect_types import resolve_type

from ntgen.config import Config
from ntgen.utils import normalize_class_name
from ntgen.utils import normalize_field_name


class NT(NamedTuple):
    """
    Contains meta data of a future NamedTuple definition.

    The NT class contains meta data which may be rendered as a single or multiple NamedTuple definitions. An instance of the class
    may be created using a direct class initialization or by parsing a dictionary using the parse_dict class method.
    """

    attrs: List[Attribute]  # type: ignore
    name: str

    def __repr__(self) -> str:
        """Format the structure name to be a Pythonic class identifier."""
        return normalize_class_name(self.name)

    @property
    def nt_attrs(self) -> List[Attribute]:
        """Return a list of all the 'user defined' type attributes of the data structure."""
        return [attr for attr in self.attrs if attr.is_user_defined]

    @property
    def builtin_type_attrs(self) -> List[Attribute]:
        """Return a list of all the builtin type attributes in the data structure."""
        return [attr for attr in self.attrs if not attr.is_user_defined]

    @classmethod
    def parse_dict(cls, data: Dict, name: str, config: Config, level: int = 0) -> Optional[NT]:
        """
        Parse a given dictionary to identify future NamedTuple metadata.

        :param data: dictionary to be parsed
        :param name: name of the future NamedTuple
        :param config: Config instance
        :param level: prevents the method from creating too deeply-nested NamedTuple definitions
        :return: An instance of the NT class if it's non-empty (i.e. it contains any attributes)
        """
        if not isinstance(data, dict):
            return None

        should_nest = config.max_level is None or level < config.max_level

        attrs = []
        for attr_name, attr_value in data.items():
            nt = NT.parse_dict(data=attr_value, name=attr_name, level=level + 1, config=config)
            attribute_type = nt if nt and should_nest else resolve_type(attr_value)

            attrs.append(Attribute(original_name=attr_name, type=attribute_type, value=attr_value))

        return NT(attrs=attrs, name=name) if attrs else None


class Attribute(NamedTuple):
    """
    A class representing metadata of a future NamedTuple definition attribute.

    The class stores metadata about a NamedTuple attribute, namely:
    - the attribute name
    - inferred attribute type
    - original value of the attribute used to infer the type

    The attribute type may be one of:
    - NoneType
    - built-in types which cannot be parameterized such as str, int, float, etc.
    - pyannotate type in case if the attribute will be rendered as a built-in parameterizable type, e.g. Dict[T,T], Tuple[T, int]
    - NT object if the attribute will be rendered as a NamedTuple definition
    """

    original_name: str
    type: Union[InternalType, NT, Type[None]]
    value: Any

    @property
    def repr_field_name(self) -> str:
        """Return the attribute name normalized to be a valid pythonic NamedTuple field name."""
        return normalize_field_name(name=self.original_name, leading_undescores_prefix=None)

    @property
    def repr_type_hint(self) -> str:
        """Return a string representing a type hint for the attribute."""
        if self.type is type(None):  # noqa: E721 # there's no other way to check for NoneType
            return "None"

        if isinstance(self.type, type):
            return self.type.__name__

        return repr(self.type)

    @property
    def is_user_defined(self) -> bool:
        """Return True if the attribute will be rendered as a user defined type (namely a NamedTuple)."""
        return isinstance(self.type, NT)



from typing import Dict
from typing import Optional

from ntgen.config import Config
from ntgen.exception import InputDataStructureIsNotADict
from ntgen.models import NT
from ntgen.render import render_nt_definition


def generate_from_dict(
    data: Dict,
    name="NTGenNamedTuple",
    snake_case: bool = True,
    camel_case: bool = True,
    insert_as_dict: bool = False,
    insert_from_dict: bool = False,
    max_level: Optional[int] = None,
) -> str:
    """
    Generate a string containing NamedTuple definition based on the given dictionary.

    :param data: dictionary to be rendered as a NamedTuple
    :param name: name for the NamedTuple
    :param snake_case: true if the NamedTuple field names should be transformed to snake_case
    :param camel_case: true if the NamedTuple class names should be transformed to CamelCase
    :param insert_as_dict: insert methods allowing for dumping nested structures to dictionaries
    :param insert_from_dict: insert generic constructors that will allow for parsing the analyzed data
    :param max_level: set the depth level for sub-tuple generation
    :return: string - valid Python code containing the definition of a NamedTuple based on the provided data structure
    """
    config = Config(
        snake_case=snake_case,
        camel_case=camel_case,
        insert_from_dict=insert_from_dict,
        insert_as_dict=insert_as_dict,
        max_level=max_level,
    )

    nt = NT.parse_dict(data=data, name=name, config=config, level=1)
    if nt is None:
        raise InputDataStructureIsNotADict(f"The input data structure was not valid. Should be a non-empty dict, was: {data}")
    return render_nt_definition(nt, config)



import pytest

from ntgen.utils import normalize_field_name


@pytest.mark.parametrize(
    ("name", "prefix", "expected_field_name"),
    (("field_name", None, "field_name"), ("__FieldName", None, "field_name"), ("___FieldName", "prefix", "prefix__field_name"),),
)
def test_normalize_field_name(name, prefix, expected_field_name):
    assert normalize_field_name(name=name, leading_undescores_prefix=prefix) == expected_field_name



import json
from typing import Optional

from tap import Tap as TypedArgumentParser

from ntgen import generate_from_dict


class ArgumentParser(TypedArgumentParser):
    """Generate NamedTuple definitions with typehints based on your data automatically."""

    input: str  # Json file containing an object with the data to analyzed
    out: Optional[str] = None  # Destination file to write the Python code to
    name: str = "NTGenTuple"  # Name of the main NamedTuple

    snake_case: bool = True  # Convert the NamedTuple field names to snake_case
    camel_case: bool = True  # Convert the NamedTuple class names to CamelCase
    constructors: bool = False  # Insert generic methods that will allow for parsing of the analyzed data structures
    as_dict: bool = False  # Insert generic methods allowing for dumping the nested NamedTuple hierarchy to a dict
    max_level: Optional[int] = None  # Specify the max nesting level of the NamedTuple

    def add_arguments(self):
        """Make the 'input' argument positional."""
        self.add_argument("input")
        self.add_argument("-s", "--snake-case", action="store_false")
        self.add_argument("-c", "--camel-case", action="store_false")
        self.add_argument("-f", "--constructors", action="store_true")
        self.add_argument("-t", "--as-dict", action="store_true")


def main() -> None:
    """Parse arguments and execute ntgen when rune form the command line interface."""
    args = ArgumentParser().parse_args()

    with open(args.input) as f:
        data = json.load(f)

    rendered_nt = generate_from_dict(
        data=data,
        name=args.name,
        snake_case=args.snake_case,
        camel_case=args.camel_case,
        insert_from_dict=args.constructors,
        insert_as_dict=args.as_dict,
        max_level=args.max_level,
    )

    if args.out is None:
        print(rendered_nt)
    else:
        with open(args.out, "w") as f:
            f.write(rendered_nt)



import os

import setuptools
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="ntgen",
    version="0.1.0",
    description=("ntgen: generate your NamedTuple definition"),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Maciej Rapacz",
    author_email="mrapacz+ntgen@protonmail.com",
    url="https://github.com/mrapacz/ntgen",
    license="MIT",
    packages=setuptools.find_packages(exclude=("tests",)),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development",
    ],
    entry_points={"console_scripts": ["ntgen=ntgen.__main__:console_entry"]},
    install_requires=["python_version >= '3.7'", "pyannotate == 1.2.0", "typed-argument-parser == 1.4",],
)



import re
from typing import Optional


def normalize_field_name(name: str, leading_undescores_prefix: Optional[str] = None) -> str:
    """
    Normalize a string to take a Pythonic form.

    Normalize a string to take a Pythonic form by:
    - replacing leading underscores with a given (optional) prefix
    - converting the name so snake_case
    """
    return convert_to_snake_case(replace_leading_underscores(name, prefix=leading_undescores_prefix))


def convert_to_snake_case(name: str) -> str:
    """
    Convert a given string to snake_case.

    Converts a given string to snake_case from camel case or kebab case
    >>> normalize_field_name('SomeCamelCase')
    'some_camel_case'
    >>> normalize_field_name('sample-kebab-case')
    'sample_kebab_case'
    """
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).replace("-", "_").lower()


def normalize_class_name(name: str) -> str:
    """
    Normalize class name by converting it to PascalCase.

    >>> normalize_class_name('some_hyphen_case')
    'SomeHyphenCase'
    """
    if re.match(r"(?:[A-Z][a-z]+)+", name):
        return name

    return "".join([fragment.capitalize() for fragment in normalize_field_name(name).split("_")])


def replace_leading_underscores(name: str, prefix=None) -> str:
    """
    Replace leading underscores with a given prefix.

    Replaces leading underscores with a given prefix. If no prefix is specified, the leading underscores are removed.
    >>> replace_leading_underscores('_private_field')
    'private_field'
    >>> replace_leading_underscores('__private_field', prefix='dunder')
    'dunder_private_field'
    """
    return re.sub(r"^_+", f"{prefix}_" if prefix else "", name)


def indent_statement(indent: int, statement: str) -> str:
    """
    Indents the given string by a specified number of indents.

    Indents the given string by a specified number of indents, e.g. indenting by 1 will preprend the string
    with 4 space characters:
    >>> indent_statement(0, 'x = 3')
    'x = 3'
    >>> indent_statement(1, 'x = 3')
    '    x = 3'
    """
    return " " * 4 * indent + statement
